/*
	Broadcasting our boids with DIS.
	Copyright (C) 2009-2010 Anthony Steed

	Based on an example from HawkNL with the following copyright:
	Copyright (C) 2000-2002 Phil Frisbie, Jr. (phil@hawksoft.com)
*/


#include <cstring>
#include <time.h>
#include <math.h>
#include <float.h>
#include "nl.h"

#if defined WIN32 || defined WIN64

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#define sleep(x)    Sleep((DWORD)(1000 * (x)))
#else
#include <unistd.h>
#define stricmp strcasecmp
#endif


#include "Vec3f.h"
#include <vector>

//
// Main parameters
//

NLint       frequency = 10;
NLsocket    sock;

//
// Flock control parameters
//

int number_boids=10;
float inertia_factor=0.1f;
float pull_factor=0.1f;
float proximity_distance=6.0f;
float proximity_factor=5.0f;
float bound_velocity=0.2f;
float max_velocity=5.0f;
Vec3f world_min_pt(-30.0f,0.0f,-30.0f);
Vec3f world_max_pt(30.0f,30.0f,30.0f);


class Boid {
public:
	Vec3f pos;
	Vec3f vel;
	float r;
	float g;
	float b;

	Boid(float red, float green, float blue) 
	{
		r = red; g = green; b = blue;
		randomPos();
	};
	void worldConstrain()
	{
		if (pos[0] > world_max_pt[0]) 
			vel[0] += -bound_velocity;
		if (pos[0] < world_min_pt[0]) 
			vel[0] +=  bound_velocity;

		if (pos[1] > world_max_pt[1]) 
			vel[1] += -bound_velocity;
		if (pos[1] < world_min_pt[1]) 
			vel[1] += bound_velocity;

		if (pos[2] > world_max_pt[2]) 
			vel[2] += -bound_velocity;
		if (pos[2] < world_min_pt[2]) 
			vel[2] +=  bound_velocity;
	};

	void randomPos()
	{
		// Initialise boid config
		pos[0] = world_min_pt[0]  + ((world_max_pt[0] - world_min_pt[0]) * (((float)rand())/RAND_MAX));
		pos[1] = world_min_pt[1]  + ((world_max_pt[1] - world_min_pt[1]) * (((float)rand())/RAND_MAX));
		pos[2] = world_min_pt[2]  + ((world_max_pt[2] - world_min_pt[2]) * (((float)rand())/RAND_MAX));
		vel = Vec3f(); // Set to null
	};
};

std::vector<Boid *> our_boids; // Our local array of boids



// Update a set of boids
void updateBoids(std::vector<Boid *> &boids, float time)
{
	int i,j;
	Vec3f center;
	Vec3f inertia;
	Vec3f pull;
	Vec3f diff;
	int proximate;

	for (i=0; i< number_boids; i++) 
	{
		center = boids[i]->pos + center; // Flock center
		inertia = boids[i]->vel + inertia; // Flock inertia
	}

	center = center * (1.0f/((float)number_boids)); // Take average to get flock center 
	inertia = inertia* (1.0f/((float)number_boids)); // Take average to get flock inertia 
	inertia = inertia *inertia_factor * time; // Scale inertia, it is the same for each boid

	for (i=0; i< number_boids; i++) 
	{	
		Vec3f local_center;

		// Flock Cohesion 
		pull = center - boids[i]->pos;
		boids[i]->vel = (pull*pull_factor*time) + boids[i]->vel;

		// Flock Alignment 
		boids[i]->vel = (inertia) + boids[i]->vel; 

		// Neighbor Separation 
		proximate = 0; // Count the number of proximal boids
		for (j=0; j< number_boids; j++) 
		{
			if (i!=j) 
			{
				diff = boids[i]->pos - boids[j]->pos;
				if (diff.length() < proximity_distance) // These are the proximal ones
				{ 
					proximate++;
					local_center = diff + local_center;
				}
			}
		}

		if (proximate>0) // If there are proximal boids
			boids[i]->vel = (local_center*proximity_factor*time) + boids[i]->vel;

		// Max velocity 
		boids[i]->vel.limit(max_velocity);

		// Calculate new position
		boids[i]->pos = boids[i]->pos + (boids[i]->vel * time);

		// World limits 
		boids[i]->worldConstrain();
	}
}


// Update a set of boids
void sendBoids(std::vector<Boid *> &boids)
{
	static NLulong timestamp=0;
	int i;
	unsigned int count;
	NLbyte buffer[1280]; /* Maximum size of a DIS PDU*/

	for (i=0; i< number_boids; i++) 
	{	
		count=0;
		// PDU Header Field
		writeByte(buffer, count, 0x06); // Protocol Version Field
		writeByte(buffer, count, 0); // Exercise Identifier Field
		writeByte(buffer, count, 0x01); // Entity State PDU
		writeByte(buffer, count, 0x01); // Entity Information/Interaction
		writeLong(buffer, count, timestamp); // Time Stamp Field
		writeShort(buffer, count, 0x0090); // PDU Length Field
		writeShort(buffer, count, 0x0000); // Padding FIeld

		// Entity Identifier Record
		writeShort(buffer, count, 0x0000); // Site Identifier Field
		writeShort(buffer, count, 0x0001); // Application Identifier Field
		writeShort(buffer, count, i+1); // Entity Identifier Field

		// Force ID Field
		writeByte(buffer, count, 0x0000); 

		// Number of Articulation Parameters Field
		writeByte(buffer, count, 0x0000); 

		// Entity Type Record
		writeByte(buffer, count, 0x0000); // Kind Field
		writeByte(buffer, count, 0x0000); // Domain Field
		writeShort(buffer, count, 0x0000); // Country Field
		writeByte(buffer, count, 0x0000); // Category Field
		writeByte(buffer, count, 0x0000); // Subcategory Field
		writeByte(buffer, count, 0x0000); // Specific Field
		writeByte(buffer, count, 0x0000); // Extra Field

		// Alternative Entity Type Record
		writeByte(buffer, count, 0x0000); // Kind Field
		writeByte(buffer, count, 0x0000); // Domain Field
		writeShort(buffer, count, 0x0000); // Country Field
		writeByte(buffer, count, 0x0000); // Category Field
		writeByte(buffer, count, 0x0000); // Subcategory Field
		writeByte(buffer, count, 0x0000); // Specific Field
		writeByte(buffer, count, 0x0000); // Extra Field

		//Entity Linear Velocity Record
		writeFloat(buffer, count, 0.0f);
		writeFloat(buffer, count, 0.0f);
		writeFloat(buffer, count, 0.0f);

		//Entity Location Record	
		// Note the rotation to DIS coordination
		writeDouble(buffer, count, boids[i]->pos[0]);
		writeDouble(buffer, count, boids[i]->pos[2]);
		writeDouble(buffer, count, -boids[i]->pos[1]);

		//Entity Orientation Record
		float rotateAngle = atan2(boids[i]->vel[0], -boids[i]->vel[2]); 
		writeFloat(buffer, count, rotateAngle);
		writeFloat(buffer, count, 0.0f);
		writeFloat(buffer, count, 0.0f);

		//Entity Appearance Record			
		writeShort(buffer, count, 0x0000); // General Appearance
		writeShort(buffer, count, 0x0000); // Specific Appearance

		// Dead Reckoning Parameter Record
		for (unsigned int k=0;k<40;k++) writeByte(buffer,count, 0x00);

		// Entity Marking Record
		writeByte(buffer, count, 0x01); // Entity Marking Character Set Field
		writeByte(buffer, count, 'B'); 
		writeByte(buffer, count, 'o'); 
		writeByte(buffer, count, 'i'); 
		writeByte(buffer, count, 'd');
		for (unsigned int k=0;k<7;k++) writeByte(buffer,count, 0x00);

		// Entity Capabilities Record
		writeLong(buffer, count, 0x00000000);
			
		//Articulation Parameter Record
		// This is optional

		(void)nlWrite(sock, (NLvoid *)buffer, (NLint)count);
	}
	timestamp++;
}


/* command line helper functions from HAWK game engine */
static int    ParamCount=0;
static char **ParamData=NULL;


static void initParam(int count, char **params)
{
	ParamCount=count;
	ParamData=params;
}

static NLboolean isParam(char *name)
{
	int i;
	int len;

	for(i=0;i<ParamCount;i++)
	{
		len = (int)strlen(ParamData[i]);
		if(stricmp(name,ParamData[i])==0)
			return NL_TRUE;
	}
	return NL_FALSE;
}

static int getIntParam(char *name)
{
	int i;

	for(i=1;i<ParamCount-1;i++) {
		if(stricmp(name,ParamData[i])==0)
			return(atoi(ParamData[i+1]));
	}
	return 0;
}

static char *getStrParam(char *name)
{
	int i;

	for(i=1;i<ParamCount-1;i++) {
		if(stricmp(name,ParamData[i])==0)
			return(ParamData[i+1]);
	}
	return NULL;
}

static float getFloatParam(char *name)
{
	int i;

	for(i=1;i<ParamCount-1;i++) {
		if(stricmp(name,ParamData[i])==0)
			return(float)(atof(ParamData[i+1]));
	}
	return 0;
}

/* main program starts here */

static void printUsage(void)
{
    printf("boid_DIS [-a multicast_address] [-p port] [-t ttl] [-f frequency]\n\n");
    printf("multicast_address full range: 224.0.0.0 to 239.255.255.255\n");
    printf("            recomended range: 234.0.0.0 to 238.255.255.255\n");
    printf("                     default: 234.235.236.237\n\n");
    printf("port number: 1024 to 65535\n");
    printf("    default: 1258\n\n");
    printf("ttl value: 1 to 255\n");
    printf("recomended values:   1 for local LAN\n");
    printf("                    15 for a site\n");
    printf("                    63 for a region\n");
    printf("                   127 for world\n");
    printf("          default: 1\n");
    printf("DO NOT use a ttl higher than you need\n\n");
    printf("frequency: 1 to 100 Hz\n");
    printf(" default: 10 Hz\n\n");
}

static void printErrorExit(void)
{
    NLenum err = nlGetError();

    if(err == NL_SYSTEM_ERROR)
    {
        printf("System error: %s\n", nlGetSystemErrorStr(nlGetSystemError()));
    }
    else
    {
        printf("HawkNL error: %s\n", nlGetErrorStr(err));
    }
    nlShutdown();
    exit(EXIT_FAILURE);
}



// update loop
void update( void )
{
    updateBoids(our_boids, 1.0f/frequency); // Updates the local boid set
    sendBoids(our_boids); // Updates the local boid set
}



int main(int argc, char **argv)
{
    NLaddress   addr, maddr;
    NLbyte      string[NL_MAX_STRING_LENGTH];
    NLbyte      address[] = "239.1.2.3";
    NLushort    port = 62040;	
    NLint       ttl = NL_TTL_LOCAL;
    NLaddress   *alladdr;
	int count;

    if(nlInit() == NL_FALSE)
        printErrorExit();

    if(nlSelectNetwork(NL_IP) == NL_FALSE)
        printErrorExit();

    initParam(argc, argv);
    if(isParam("-h") == NL_TRUE)
    {
        printUsage();
        return 0;
    }
    if(isParam("-a") == NL_TRUE)
    {
        nlStringToAddr(getStrParam("-a"), &maddr);
    }
    else
    {
        nlStringToAddr(address, &maddr);
    }
    if(isParam("-p") == NL_TRUE)
    {
        NLushort p = (NLushort)getIntParam("-p");

        if(p < 1024 || p > 65535)
        {
            printf("error: invalid port, using default port of %u\n", port);
            nlSetAddrPort(&maddr, port);
        }
        else
        {
            nlSetAddrPort(&maddr, p);
        }
    }
    else
    {
        nlSetAddrPort(&maddr, port);
    }

    if(isParam("-t") == NL_TRUE)
    {
        NLint t = getIntParam("-t");

        if(t < 1 || t > 255)
        {
            printf("error: invalid ttl, using default ttl of %d\n", ttl);
        }
        else
        {
            ttl = t;
        }
    }
    if(isParam("-f") == NL_TRUE)
    {
        NLint i = getIntParam("-i");

        if(i < 1 || i > 60)
        {
            printf("error: invalid frequency, using default frequency of %d\n", frequency);
        }
        else
        {
            frequency = i;
        }
    }

	/* change TTL if you want to get out of your LAN */
    /* note that this call must be before nlConnect */
    nlHint(NL_MULTICAST_TTL, ttl);

	/* enable reuse address to run two or more copies on one machine */
    nlHint(NL_REUSE_ADDRESS, NL_TRUE);

	/* list all the local addresses */
    printf("local addresses are:\n");
    alladdr = nlGetAllLocalAddr(&count);
    if(alladdr != NULL)
    {
        while(count-- > 0)
        {
            printf("  %s\n", nlAddrToString(alladdr++, string));
        }
    }
    printf("\n");

	sock = nlOpen(0, NL_UDP_MULTICAST);

	if(sock == NL_INVALID)
		printErrorExit();

	nlGetLocalAddr(sock, &addr);
	printf("our socket address is %s\n", nlAddrToString(&addr, string));

	if(nlConnect(sock, &maddr) == NL_FALSE)
	{
		nlClose(sock);
		printErrorExit();
	}
	printf("multicast address %s\n", nlAddrToString(&maddr, string));

	for (int i=0;i<number_boids;i++) // Create a flock
	{
		our_boids.push_back(new Boid(0.0f, 0.0f, 1.0f));
	}

	while(1 == 1)
	{
		update();
		sleep(1.0/frequency);
	}

	return 0;
}

